#!/bin/node

const hk = require('./hikvision_dl');
const tspd = require('parallel_dl');
const url = require('url');
const fs = require('fs');
const promisify = require('util').promisify;

function StrToDate(strdate){
    //WARNING: User-friendly, but not (NOOB||HACKER)-checking
    //Please do not pentest this function
    //User-friendly date string
//     d.replace(/ /gi, "");
    d = strdate.split(".");
    
    var r = new Date(0);
    console.log("d: ", d);
    r.setDate(parseInt(d[0]));
    r.setMonth(parseInt(d[1]) - 1);
    var sp = d[2].replace(/( )+/gi, " ").split(" ");
    r.setYear(sp[0]);
    if(sp.length > 1){
        var tsp = sp[1].split(":"); //Time split
        r.setHours(tsp[0]);
        r.setMinutes(tsp[1]);
        r.setSeconds(tsp[2] || 0);
    }
    return r;
}

let args = process.argv;
// if(args[0].split(/\\|\//).splice(-1) === 'node'){
//
// }
args.splice(0, 2);

function help(){
    console.log(`Usage:
    hikdl "http://HOST:PORT" [login] [password] [channel] [from_date] [to_date] [output_dir]
        Date format: '1.7.2022  13:30'  (write your own)
    Example:
        hikdl https://tssoftware.eu:6000 admin password123 201 '5.12.2020    8:30' '6.12.2020  17:55' ~/Videos/Camera
                        [url]           [user]  [passwd] [channel]      [from-date]         [to-date]      [output folder]   - Will download all clips between the specified times
        OR
            hikdl [url] [user] [pass] [command]
        
        COMMANDS:
            channels    | List channels
            tryauth | Only try to log in
        If you want more automation, feel free to use hikvison_dl as library for your own script

    `)
    process.exit(1);
}


async function auth(client, proto = "http"){
    if(!(await client.login(proto))[0]){
        console.log("ERROR: Cannot log-in - bad password or not compatible");
        process.exit(5);
    }
}

(async function(){
    let urlp = url.parse(args[0]);
    client = new hk.HikvisionClient(urlp.hostname, urlp.port, args[1], args[2]);
    if((args.length < 4) || (args[0] == "--help")){
        help();
        return;
    }

    if(args.length < 7){
        switch(args[3][0]){
            case "t":
                let res = (await client.login("http"))[0];
                console.log(`res: ${res}`);
                console.log((res) ? "Success" : "Cannot log in");
                process.exit((res) ? 0 : 2)
                return;
            case "c":
                await auth(client, urlp.proto);
                let channels = await client.getChannels();
                if(args[4] == "json"){
                    console.log(JSON.stringify(channels));
                }else{
                    for(i in channels){
                        let c = channels[i];
                        process.stdout.write(`${c.id}|${c.name}|${c.sourceInputPorts}`);
                    }
                }
                break;
            default:
                console.error(`ERROR: Invalid command: ${args[3]}`);
                process.exit(1);
        }
        process.exit(0);
    }else{
        await auth(client, urlp.proto);
    }

    //Download videos (default action)    
    var pd = new tspd.ParallelDownloader();
    let fls = await client.getFilesBetween_pdl(args[3], StrToDate(args[4]), StrToDate(args[5]))
    if(fls === false){
        console.error("ERROR: Unable to get tracks in that time range");
        process.exitCode = 5;
        return;
    }
//     let fls = await client.getFilesBetween_pdl("101", StrToDate("1.7.2022   13:30"), StrToDate("1.7.2022   18:30"))
    console.log("fls: ", fls);
    try{
        await promisify(fs.mkdir)(args[6]);
    }catch(e){
        if(e.code !== 'EEXIST'){
            throw e;
        }
    }
    pd.addUrls(fls, {outputDirectory: args[6]});
    await pd.downloadCLI(2);
})()
